{=========================================================================
  IntList Unit

  This unit provides a TIntegerList class that mimics the TStringList
  class except that this class manages a list of Integers.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit IntList;

interface

uses
  Classes;

type
  EOutOfRange   = class( EListError );

  TIntegerList = class( TPersistent )
  private
    FList : TList;
    FDuplicates : TDuplicates;
    FMin : Integer;
    FMax : Integer;
    FSorted : Boolean;
    function GetCount : Integer;
    procedure ReadMin( Reader : TReader );
    procedure WriteMin( Writer : TWriter );
    procedure ReadMax( Reader : TReader );
    procedure WriteMax( Writer : TWriter );
    procedure ReadIntegers( Reader : TReader );
    procedure WriteIntegers( Writer : TWriter );
    procedure SetMin( Value : Integer );
    procedure SetMax( Value : Integer );
    procedure SetSorted( Value : Boolean );
    procedure QuickSort( L, R : Integer );
  protected
    procedure DefineProperties( Filer : TFiler ); override;
    function Find( N : Integer; var Index : Integer ) : Boolean; virtual;
    function GetItem( Index : Integer ) : Integer; virtual;
    procedure SetItem( Index : Integer; Value : Integer ); virtual;
    procedure Sort; virtual;
  public
    constructor Create;
    destructor Destroy; override;

    function Add( Value : Integer ) : Integer; virtual;
    procedure AddIntegers( List : TIntegerList ); virtual;
    procedure Assign( Source : TPersistent ); override;
    procedure Clear; virtual;
    procedure Delete( Index : Integer ); virtual;
    function Equals( List : TIntegerList ) : Boolean;
    procedure Exchange( Index1, Index2 : Integer); virtual;
    function IndexOf( N : Integer ) : Integer; virtual;
    procedure Insert( Index : Integer; Value : Integer ); virtual;
    procedure Move( CurIndex, NewIndex : Integer ); virtual;

    property Duplicates : TDuplicates
      read FDuplicates
      write FDuplicates;

    property Count : Integer
      read GetCount;

    property Items[ Index : Integer ] : Integer
      read GetItem
      write SetItem; default;

    property Min : Integer
      read FMin
      write SetMin;

    property Max : Integer
      read FMax
      write SetMax;

    property Sorted : Boolean
      read FSorted
      write SetSorted;
  end;



implementation

uses
  WinTypes;

{==========================}
{== TIntegerList Methods ==}
{==========================}

constructor TIntegerList.Create;
begin
  inherited Create;
  FList := TList.Create;          // Create internal list to hold integers
end;


destructor TIntegerList.Destroy;
begin
  Clear;              // Ensure memory allocated for each item is released
  FList.Free;                        // Don't forget to free the container
  inherited Destroy;
end;


procedure TIntegerList.Assign( Source : TPersistent );
begin
  if Source is TIntegerList then
  begin
    Clear;
    AddIntegers( TIntegerList( Source ) );
  end
  else
    inherited Assign( Source );
end;


{=========================================================================
  TIntegerList.DefineProperties

  This method instructs Delphi to include the Min and Max property values
  whenever a TIntegerList object is read from or written to a stream
  (e.g. a form file).  This method also ensures that the list of integers
  gets stored in the stream as well.
=========================================================================}

procedure TIntegerList.DefineProperties( Filer : TFiler );

  function DoWrite : Boolean;
  begin
    if Filer.Ancestor <> nil then
    begin
      Result := True;
      // If form inheritance is being used, then only store integers
      // if they differ from the ancestor property list.
      if Filer.Ancestor is TIntegerList then
        Result := not Equals( TIntegerList( Filer.Ancestor ) );
    end
    else Result := Count > 0;
  end;

begin
  Filer.DefineProperty( 'Min', ReadMin, WriteMin, FMin <> 0 );
  Filer.DefineProperty( 'Max', ReadMax, WriteMax, FMax <> 0 );
  Filer.DefineProperty( 'Integers', ReadIntegers, WriteIntegers, DoWrite);
end;


procedure TIntegerList.ReadMin( Reader : TReader );
begin
  FMin := Reader.ReadInteger;
end;


procedure TIntegerList.WriteMin( Writer : TWriter );
begin
  Writer.WriteInteger( FMin );
end;


procedure TIntegerList.ReadMax( Reader : TReader );
begin
  FMax := Reader.ReadInteger;
end;


procedure TIntegerList.WriteMax( Writer : TWriter );
begin
  Writer.WriteInteger( FMax );
end;


procedure TIntegerList.ReadIntegers( Reader : TReader );
begin
  Reader.ReadListBegin;                // Read in the start of list marker
  Clear;                                             // Clear current list
  while not Reader.EndOfList do
    Add( Reader.ReadInteger );              // Add stored integers to list
  Reader.ReadListEnd;                    // Read in the end of list marker
end;


procedure TIntegerList.WriteIntegers( Writer : TWriter );
var
  I : Integer;
begin
  Writer.WriteListBegin;      // Be sure to store the start of list marker
  for I := 0 to Count - 1 do
    Writer.WriteInteger( GetItem( I ) );    // Send all integers to Writer
  Writer.WriteListEnd;                     // Store the end of list marker
end;


procedure TIntegerList.SetSorted( Value : Boolean );
begin
  if FSorted <> Value then
  begin
    if Value then
      Sort;
    FSorted := Value;
  end;
end;


function TIntegerList.GetCount : Integer;
begin
  Result := FList.Count;
end;


function TIntegerList.GetItem( Index : Integer ) : Integer;
begin
  Result := PInteger( FList.Items[ Index ] )^;
end;


procedure TIntegerList.SetItem( Index : Integer; Value : Integer );
begin
  if ( FMin <> FMax ) and ( ( Value < FMin ) or ( Value > FMax ) ) then
    raise EOutOfRange.CreateFmt( 'Value must be within %d..%d',
                                 [ FMin, FMax ] );

  PInteger( FList.Items[ Index ] )^ := Value;
end;


procedure TIntegerList.SetMin( Value : Integer );
var
  I : Integer;
begin
  if Value <> FMin then
  begin
    // Scan list to make sure new min value is valid
    for I := 0 to Count - 1 do
    begin
      if GetItem( I ) < Value then
        raise EOutOfRange.CreateFmt( 'Unable to set new minimum ' +
                                     'value.'#13'List contains values ' +
                                     'below %d',[ Value ] );
    end;
    FMin := Value;
    if FMin > FMax then
      FMax := FMin;
  end;
end; {= TIntegerList.SetMin =}


procedure TIntegerList.SetMax( Value : Integer );
var
  I : Integer;
begin
  if Value <> FMax then
  begin
    for I := 0 to Count - 1 do
    begin
      if GetItem( I ) > Value then
        raise EOutOfRange.CreateFmt( 'Unable to set new maximum ' +
                                     'value.'#13'List contains values ' +
                                     'above %d',[ Value ] );
    end;
    FMax := Value;
    if FMax < FMin then
      FMin := FMax;
  end;
end; {= TIntegerList.SetMax =}


procedure TIntegerList.AddIntegers( List : TIntegerList );
var
  I : Integer;
begin
  for I := 0 to Pred( List.Count ) do
    Add( List[ I ] );
end;


function TIntegerList.Add( Value : Integer ) : Integer;
begin
  Result := Count;
  Insert( Count, Value );
end;


procedure TIntegerList.Clear;
var
  I : Integer;
begin
  for I := 0 to Pred( FList.Count ) do
    Dispose( PInteger( FList.Items[ I ] ) );
  FList.Clear;
end;


procedure TIntegerList.Delete( Index : Integer );
begin
  Dispose( PInteger( FList.Items[ Index ] ) );
  FList.Delete( Index );
end;


function TIntegerList.Equals( List : TIntegerList ) : Boolean;
var
  I, Count : Integer;
begin
  Count := GetCount;
  if Count <> List.GetCount then
    Result := False
  else
  begin
    I := 0;
    while ( I < Count ) and ( GetItem( I ) = List.GetItem( I ) ) do
      Inc( I );
    Result := I = Count;
  end;
end; {= TIntegerList.Equals =}


procedure TIntegerList.Exchange( Index1, Index2 : Integer );
begin
  FList.Exchange( Index1, Index2 );
end;


{=========================================================================
  Find

  A protected method that implements a binary search function that is
  called by IndexOf only if the list is sorted.
=========================================================================}

function TIntegerList.Find( N : Integer; var Index : Integer ) : Boolean;
var
  L, H, I : Integer;
begin
  Result := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := ( L + H ) shr 1;
    if PInteger( FList[ I ] )^ < N then
      L := I + 1
    else
    begin
      H := I - 1;
      if PInteger( FList[ I ] )^ = N then
      begin
        Result := True;
        if Duplicates <> dupAccept then
          L := I;
      end;
    end;
  end;
  Index := L;
end; {= TIntegerList.Find =}


function TIntegerList.IndexOf( N : Integer ) : Integer;
var
  I : Integer;
begin
  Result := -1;

  if not Sorted then
  begin
    I := 0;
    while ( I < GetCount ) and ( GetItem( I ) <> N ) do
      Inc( I );
    if I <> GetCount then
      Result := I;
  end
  else if Find( N, I ) then                // If list is sorted, call Find
    Result := I;
end; {= TIntegerList.IndexOf =}


procedure TIntegerList.Insert( Index : Integer; Value : Integer );
var
  P : PInteger;
begin
  if ( FMin <> FMax ) and ( ( Value < FMin ) or ( Value > FMax ) ) then
    raise EOutOfRange.CreateFmt( 'Value must be within %d..%d',
                                 [ FMin, FMax ] );

  New( P );                              // Allocate memory for an integer
  P^ := Value;
  FList.Insert( Index, P );           // Insert integer onto internal list
end;


procedure TIntegerList.Move( CurIndex, NewIndex : Integer );
begin
  FList.Move( CurIndex, NewIndex );
end;


procedure TIntegerList.QuickSort( L, R : Integer );
var
  I, J : Integer;
  P    : PInteger;
begin                                       // Generic QuickSort Procedure
  I := L;
  J := R;
  P := PInteger( FList[ ( L + R ) shr 1 ] );
  repeat
    while PInteger( FList[ I ] )^ < P^  do
      Inc( I );
    while PInteger( FList[ J ] )^ > P^ do
      Dec( J );
    if I <= J then
    begin
      FList.Exchange( I, J );
      Inc( I );
      Dec( J );
    end;
  until I > J;
  if L < J then
    QuickSort( L, J );
  if I < R then
    QuickSort( I, R );
end; {= TIntegerList.QuickSort =}


procedure TIntegerList.Sort;
begin
  if not Sorted and ( FList.Count > 1 ) then
    QuickSort( 0, FList.Count - 1 );
end;

end.